package ru.ifmo.image;

import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.imageio.ImageIO;

/**
 * ,      .
 * 
 *        .
 *  ,     , 
 *   .    
 *   {@link #paint(Graphics2D)}.
 * 
 * @author Egorov Kirill
 */
public class RotatableImage {

	/**
	 *  .
	 */
	protected BufferedImage image = null;

	/**
	 *  .
	 */
	protected int width = 0;

	/**
	 *  .
	 */
	protected int height = 0;
	
	/**
	 * x- .
	 */
	protected volatile double x = 0;

	/**
	 * y- .
	 */
	protected volatile double y = 0;
	
	/**
	 *   ,    x.
	 */
	protected volatile double angle = 0;

	/**
	 *  ,   .
	 */
	protected final Component drawPane;

	/**
	 *      .
	 */
	protected AffineTransform transform = new AffineTransform();
	
	/**
	 *  lock    
	 */
	protected final ReentrantReadWriteLock rwLock = new ReentrantReadWriteLock();
	protected final Lock rLock = rwLock.readLock();
	protected final Lock wLock = rwLock.writeLock();

	private Point2D[] rectSrc = new Point2D.Double[4];
	private Point2D[] rectDst = new Point2D.Double[4];
	
	/**
	 * ,    <code>RotatableImage</code>.
	 *    {@link #setImage(BufferedImage)}.
	 * @param parent     
	 */
	public RotatableImage(Component parent) {
		drawPane = parent;
		for (int i = 0; i < rectSrc.length; i++) {
			rectSrc[i] = new Point2D.Double();
			rectDst[i] = new Point2D.Double();
		}
	}
	
	/**
	 * ,    <code>RotatableImage</code>  , 
	 *    <code>filename</code>.
	 * @param parent     
	 * @param name    
	 * @throws IOException 
	 */
	public RotatableImage(Component parent, String filename) throws IOException {
		this(parent);
		
		setImage(loadImage(parent, filename));
	}
	
	/**
	 *    <code>filename</code>
	 * @param parent  
	 * @param filename    
	 * @return  
	 * @throws IOException
	 */
	public static BufferedImage loadImage(Component parent, String filename)  throws IOException {
		URLClassLoader urlLoader = (URLClassLoader)(
				(parent != null) ? parent.getClass().getClassLoader() : ClassLoader.getSystemClassLoader());
		URL fileLoc = urlLoader.findResource(filename);
		BufferedImage image = ImageIO.read(fileLoc);
		if (image == null) {
			throw new IOException("Cannot read " + filename);
		}
		return image;
	}
	
	/**
	 *   .
	 * @param img 
	 */
	public void setImage(BufferedImage img) {
		wLock.lock();
		try {
			image = img;
			
			width = image.getWidth();
			height = image.getHeight();
			
			rectSrc[0].setLocation(0, 0);
			rectSrc[1].setLocation(0, height);
			rectSrc[2].setLocation(width, height);
			rectSrc[3].setLocation(width, 0);
		} finally{
			wLock.unlock();
		}
	}
	
	/**
     * Stores the x,y origin of the object(image) into "return value"
     * <code>rv</code> and returns <code>rv</code>.
     * If <code>rv</code> is <code>null</code> a new <code>Point</code>
     * is allocated.  This version of <code>getLocation</code> is useful
     * if the caller wants to avoid allocating a new <code>Point</code>
     * object on the heap.
     *
     * @param rv the return value, modified to the objects's location
     * @return <code>rv</code>
     */
	public Point2D getLocation(Point2D rv) {
		rLock.lock();
		try {
			if (rv == null) {
	            return new Point2D.Double(x, y);
	        }
	        else {
	            rv.setLocation(x, y);
	            return rv;
	        } 
		} finally {
			rLock.unlock();
		}
	}
	
	/**
	 * x-  .
	 * @return x- 
	 */
	public double getX() {
		return x;
	}
	
	/**
	 * y-  .
	 * @return y- 
	 */
	public double getY() {
		return y;
	}
	
	/**
	 *    .
	 * @return   
	 */
	public double getDirectionAngle() {
		return angle;
	}
	
	/**
	 *   .
	 * @return 
	 */
	public int getHeight() {
		return height;
	}
	
	/**
	 *   .
	 * @return 
	 */
	public int getWidth() {
		return width;
	}
	
	/**
	 *     (x,y).
	 * @param x    x 
	 * @param y    y
	 */
	public void setLocation(double x, double y) {
		wLock.lock();
		try {
			this.x = x;
			this.y = y;
		} finally {
			wLock.unlock();
		}
	}
	
	/**
	 *     <code>p</code>.
	 * @param p   
	 */
	public void setLocation(Point2D p) {
		wLock.lock();
		try {
			x = p.getX();
			y = p.getY();
		} finally {
			wLock.unlock();
		}	
	}
	
	/**
	 *      <code>length</code>, 
	 *    .
	 * @param length ,     
	 */
	public void move(double length) {
		wLock.lock();
		try {
			x += length * Math.cos(angle);
			y += length * Math.sin(angle);
		} finally {
			wLock.unlock();
		}			
	}
	
	/**
	 *     <code>theta</code>. 
	 * @param theta   
	 */
	public void rotate(double theta) {
		wLock.lock();
		try {
			angle += theta;
			angle = Math.IEEEremainder(angle, Math.PI *2);
		} finally {
			wLock.unlock();
		}		
	}
	
	/**
	 *     <code>theta</code>. 
	 * @param theta   
	 */
	public void setDirectionAngle(double theta) {
		wLock.lock();
		try {
			angle = Math.IEEEremainder(theta, Math.PI *2);
		} finally {
			wLock.unlock();
		}		
	}
	
	/**
	 * 
	 * @param g
	 * @deprecated
	 */
	public void clearDrawArea(Graphics2D g) {
		rLock.lock();
		try {
			transform.transform(rectSrc, 0, rectDst, 0, rectSrc.length);
			double maxX = rectDst[0].getX();
			double maxY = rectDst[0].getY();
			double minX = rectDst[0].getX();
			double minY = rectDst[0].getY();
			for (Point2D p: rectDst) {
				double x = p.getX();
				if (maxX < x) maxX = x;
				else if (minX > x) minX = x;
				double y = p.getY();
				if (maxY < y) maxY = y;
				else if (minY > y) minY = y; 
			}
			g.clearRect((int)minX - 1, (int)minY - 1, (int)(maxX - minX) + 2, (int)(maxY - minY) + 2);
		} finally {
			rLock.unlock();
		}	
		
	}

	/**
	 *     .
	 * @param g <code>Graphics2D</code>  
	 */
	public void paint(Graphics2D g) {
		transform();
		g.drawImage(image, transform, drawPane);
	}
	
	/**
	 *        
	 *     {@link #angle}  .
	 */
	protected void transform() {
		rLock.lock();
		try {
			transform.setToTranslation(x - width / 2, y - height / 2);
			transform.rotate(angle, width / 2, height / 2);
		} finally {
			rLock.unlock();
		}			
	}
	
	public String toString() {
		String str;
		rLock.lock();
		try {
			str = String.format("(%f, %f) angle = %f", x, y, angle);
		} finally {
			rLock.unlock();
		}		
		return str; 
	}
}
